library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

library UNISIM;
use UNISIM.VComponents.all;


entity GTY_Transceiver_TB is
--  Port ( );
end GTY_Transceiver_TB;

architecture Behavioral of GTY_Transceiver_TB is
subtype sl is std_logic;
subtype slv is std_logic_vector;

constant  Period_Free : time := 10 ns; -- 100 MHz
constant  Period_Ref  : time := 5 ns; -- 200 MHz
signal    Clk_Free    : sl;
signal    Clk_Ref     : sl;

---------------------------------------------------------------------------------------------
-- Channel Bondign Disabled (Is used in Multi-Lane Configuration)
-- If Clock Correction is not used, the rxusrclk and txusrclock
-- Must be sourced from the same BUFG_GT! Not used in this TB.
-- More on clock Corrrection: UG578 / P283 "Setting Clock Correction Sequences"
-- RX_DISPERR_SEQ_MATCH (Channel Bonding and Clock Correction bust be matched)
--
-- Comma:ALIGN_COMMA_DOUBLE(false!):
-- Can be used only for 16/20 Width ==> ALIGN_PCOMMA_DET == ALIGN_MCOMMA_DET
-- Can be used only for 16/20 Width ==> RXPCOMMAALIGNEN  == RXMCOMMAALIGNEN
-- 
-- Plus Comma  : 1100000101 == Positive RD K28.5
-- Minus Comma : 0011111010 == Negative RD K28.5
-- Alignment Boundary : 4 Bytes for 32/40 Serdes Width
-- 
-- Running Disparity "RD" : Positive        Negative
-- K28.5 : [1011_1100]     [ 11_0000_0101]  [00_1111_1010]
-- TXCTRL1 & TXCTRL0:
--     0       0 --> Encoder Chooses Disparity
--     0       1 --> Inverts Disparity
--     1       0 --> Force Negative Disparity
--     1       1 --> Force Positive Disparity
--
-- Clock Correction :1 Sequence, 4 Words:
-- See UG578 "Valid ALIGN_COMMA_WORD/CLK_COR_SEQ_LEN Combinations"
-- 11110111 "K" == K_23_7
-- 11110111 "K" == K_23_7
-- 11110111 "K" == K_23_7
-- 11110111 "K" == K_23_7
-- If clock correction is used, RXUSRCLK and RXUSRCLK2 can be sourced by RXOUTCLK or TXOUTCLK
---------------------------------------------------------------------------------------------
component Aurora_GTY is 
port(
  gtwiz_userclk_tx_active_in          : in  slv(0 downto 0);
  gtwiz_userclk_rx_active_in          : in  slv(0 downto 0);
  gtwiz_reset_clk_freerun_in          : in  slv(0 downto 0);
  gtwiz_reset_all_in                  : in  slv(0 downto 0);
  gtwiz_reset_tx_pll_and_datapath_in  : in  slv(0 downto 0);
  gtwiz_reset_tx_datapath_in          : in  slv(0 downto 0);
  gtwiz_reset_rx_pll_and_datapath_in  : in  slv(0 downto 0);
  gtwiz_reset_rx_datapath_in          : in  slv(0 downto 0);
  gtwiz_reset_rx_cdr_stable_out       : out slv(0 downto 0);
  gtwiz_reset_tx_done_out             : out slv(0 downto 0);
  gtwiz_reset_rx_done_out             : out slv(0 downto 0);

  -------------------------------
  ------ SERDES DATA PORTS ------
  gtwiz_userdata_tx_in                : in slv(31 downto 0);
  gtwiz_userdata_rx_out               : out slv(31 downto 0);
  
  gtrefclk00_in         : in slv(0 downto 0);
  gtrefclk01_in         : in slv(0 downto 0);
  qpll0outclk_out       : out slv(0 downto 0);
  qpll0outrefclk_out    : out slv(0 downto 0);
  qpll1outclk_out       : out slv(0 downto 0);
  qpll1outrefclk_out    : out slv(0 downto 0);
  
  -----------------------------------------
  ------- Custom Functionality Setup ------
  loopback_in           : in slv(2 downto 0);
  rx8b10ben_in          : in slv(0 downto 0);
  rxpd_in               : in slv(1 downto 0);
  txpd_in               : in slv(1 downto 0);
  rxprbssel_in          : in slv(3 downto 0);
  txprbssel_in          : in slv(3 downto 0);
  tx8b10ben_in          : in slv(0 downto 0);
  rxpolarity_in         : in slv(0 downto 0);
  txpolarity_in         : in slv(0 downto 0);
  rxpmareset_in         : in slv(0 downto 0);
  txpmareset_in         : in slv(0 downto 0);

  ------------------------------------
  -------- Driver Configuration ------
  txprecursor_in        : in slv(4 downto 0);
  txpostcursor_in       : in slv(4 downto 0);
  txdiffctrl_in         : in slv(4 downto 0);

  ------------------------------------
  ------- PRBS Status / Control ------
  rxprbserr_out         : out slv(0 downto 0);
  rxprbslocked_out      : out slv(0 downto 0);
  rxprbscntreset_in     : in  slv(0 downto 0);

  ----------------------------
  ------- User Clocking ------
  rxusrclk_in           : in slv(0 downto 0);
  rxusrclk2_in          : in slv(0 downto 0);
  txusrclk_in           : in slv(0 downto 0);
  txusrclk2_in          : in slv(0 downto 0);

  -----------------------------------
  ------- Serdes Differential ------- 
  gtyrxn_in             : in  slv(0 downto 0);
  gtyrxp_in             : in  slv(0 downto 0);
  gtytxn_out            : out slv(0 downto 0);
  gtytxp_out            : out slv(0 downto 0);

  ---------------------------------
  ------- RX / TX Registers -------
  rxctrl0_out           : out slv(15 downto 0);
  rxctrl1_out           : out slv(15 downto 0);
  rxctrl2_out           : out slv(7 downto 0);
  rxctrl3_out           : out slv(7 downto 0);
  txctrl0_in            : in slv(15 downto 0);
  txctrl1_in            : in slv(15 downto 0);
  txctrl2_in            : in slv(7 downto 0);

  -------------------------------------------------
  ------- Aurora "Comma Alignment" 8/10 Ports -----
  rxcommadeten_in       : in slv(0 downto 0);
  rxmcommaalignen_in    : in slv(0 downto 0);
  rxpcommaalignen_in    : in slv(0 downto 0);

  rxbyteisaligned_out   : out slv(0 downto 0);
  rxbyterealign_out     : out slv(0 downto 0);
  rxcommadet_out        : out slv(0 downto 0);

  -------------------------------------
  ------- Clock Correction Ports ------
  rxbufstatus_out       : out slv(2 downto 0);
  rxbufreset_in         : in  slv(0 downto 0);
  rxclkcorcnt_out       : out slv(1 downto 0);
  
  ----------------------------------
  ------- RX and TX Clocking -------
  rxoutclk_out          : out slv(0 downto 0);
  txoutclk_out          : out slv(0 downto 0);

  gtpowergood_out       : out slv(0 downto 0);
  rxpmaresetdone_out    : out slv(0 downto 0);
  txpmaresetdone_out    : out slv(0 downto 0)
);
end component;

--------------------------------------------
------- PowerUp/Reset/Status Signals -------
signal Transceiver_Reset_TX : sl := '0';
signal Transceiver_Reset_RX : sl := '0';
signal Transceiver_Reset    : sl := '0';
signal GT_PowerGood         : sl := '0';
signal CDR_Stable           : sl := '0';
signal RX_PMA_Rst_Done      : sl := '0';
signal TX_PMA_Rst_Done      : sl := '0';
signal GT_TX_Reset_Done     : sl := '0';
signal GT_RX_Reset_Done     : sl := '0';
signal RX_PMA_Reset         : sl := '0';
signal TX_PMA_Reset         : sl := '0';

signal TX_Clk_Raw         : sl := '0';
signal TX_Clk_Buffed      : sl := '0';
signal RX_Clk_Raw         : sl := '0';
signal RX_Clk_Buffed      : sl := '0';
signal TX_Active          : sl := '0';
signal RX_Active          : sl := '0';
signal TX_Clk_Active_Meta : sl := '0';
signal TX_Clk_Active_Sync : sl := '0';
signal RX_Clk_Active_Meta : sl := '0';
signal RX_Clk_Active_Sync : sl := '0';
signal Serdes_Wire_N      : sl := '0';
signal Serdes_Wire_P      : sl := '0';

signal TX_Data            : slv(31 downto 0)  := (others => '0');
signal RX_Data            : slv(31 downto 0)  := (others => '0');
signal TX_CharIS_K        : slv(3 downto 0)   := "0000";
signal TX_Disp_0          : slv(3 downto 0)   := "0000";
signal TX_Disp_1          : slv(3 downto 0)   := "0000";
signal RX_CharIS_K        : slv(3 downto 0)   := "0000";
signal RX_Disp_Err        : slv(3 downto 0)   := "0000";
signal RX_CharIS_Comma    : slv(3 downto 0)   := "0000";
signal RX_NotInTable      : slv(3 downto 0)   := "0000";

type Transceiver_Reset_t is (WAIT_POWER_GOOD,WAIT_TX_RESET_COMPLETE,WAIT_RX_RESET_COMPLETE,RESET_DONE);
signal AuroraReset_FSM : Transceiver_Reset_t;


----------------------------------
------ Polarity Inversions -------
-- Might Require to Enable "Additional Ports in GTWIZ"
constant POLARITY_NORMAL    : sl := '0';
constant POLARITY_INVERTED  : sl := '1';

----------------------------------------
------- 8/10 Encoding / Decoding  ------
-- Might Require to Enable "Additional Ports in GTWIZ"
constant B8B10_ENABLED      : sl := '1';
constant B8B10_DISABLED     : sl := '0';

-----------------------------------------------
------- Synchronization Comma ALignment -------
-- Might Require to Enable "Additional Ports in GTWIZ"
constant COMMA_ALIGN_ENABLED  : sl := '1';
constant COMMA_ALIGN_DISABLED : sl := '0';

-------------------------------------------------------
------- Pseudorandom Sequence Generation Option -------
-- Might Require to Enable "Additional Ports in GTWIZ""
constant PRBS_NORMAL    : slv(3 downto 0) := "0000";
constant PRBS_7         : slv(3 downto 0) := "0001";
constant PRBS_9         : slv(3 downto 0) := "0010";
constant PRBS_15        : slv(3 downto 0) := "0011";
constant PRBS_23        : slv(3 downto 0) := "0100";
constant PRBS_31        : slv(3 downto 0) := "0101";
signal PRBS_Mode        : slv(3 downto 0) := PRBS_NORMAL;
signal PRBS_Error       : sl := '0';
signal PRBS_Locked      : sl := '0';
signal PRBS_Reset       : sl := '0';

-----------------------------------------------------------------
------- PowerDown Options (In Case only RX / TX  is Used) -------
-- Might Require to Enable "Additional Ports in GTWIZ"
constant PD_NORMAL      : slv(1 downto 0) := "00"; -- Normal Operation
constant PD_REC_FAST    : slv(1 downto 0) := "01"; -- Powerdown - Fast Recovery
constant PD_REC_SLOW    : slv(1 downto 0) := "10"; -- Powerdown - Slow Recovery
constant PD_POWERDOWN   : slv(1 downto 0) := "11"; -- Complete Powerdown

--------------------------------
------- LoopBack Options -------
-- PCS == Physical Coding Sublayer
-- PMA == Physical Medium Attachement
-- Far-End PCS  ( RX --> TX )
-- Far-End PMA  ( RX --> TX )
-- Near-End PCS ( TX --> RX ) 
-- Near-End PMA ( TX --> RX )
constant LOOP_NORMAL    : slv(2 downto 0) := "000";
constant LOOP_NEAR_PCS  : slv(2 downto 0) := "001";
constant LOOP_NEAR_PMA  : slv(2 downto 0) := "010";
constant LOOP_FAR_PMA   : slv(2 downto 0) := "100";
constant LOOP_FAR_PCS   : slv(2 downto 0) := "110";
signal Loopback_Mode    : slv(2 downto 0) := LOOP_NORMAL;

-------------------------------
------ RX Buffer Status -------
-- Clock Correction must be enabled
constant RXBUF_NORMAL     : slv(2 downto 0) := "000";
constant RXBUF_LESS       : slv(2 downto 0) := "001";
constant RXBUF_MORE       : slv(2 downto 0) := "010";
constant RXBUF_UNDERFLOW  : slv(2 downto 0) := "101";
constant RXBUF_OVERFLOW   : slv(2 downto 0) := "110";

-- UG578 / 576 "Valid Control K Characters"
-- Comma symbols: K.28.1, K.28.5, and K.28.
constant K_28_0 : slv(7 downto 0) := "00011100";
constant K_28_1 : slv(7 downto 0) := "00111100";
constant K_28_2 : slv(7 downto 0) := "01011100";
constant K_28_3 : slv(7 downto 0) := "01111100";
constant K_28_4 : slv(7 downto 0) := "10011100";
constant K_28_5 : slv(7 downto 0) := "10111100";
constant K_28_6 : slv(7 downto 0) := "11011100";
constant K_28_7 : slv(7 downto 0) := "11111100";
constant K_23_7 : slv(7 downto 0) := "11110111";
constant K_27_7 : slv(7 downto 0) := "11111011";
constant K_29_7 : slv(7 downto 0) := "11111101";
constant K_30_7 : slv(7 downto 0) := "11111110";
constant D_10_2 : slv(7 downto 0) := "01001010";
constant D_21_5 : slv(7 downto 0) := "10110101";
-- Aurora 8/10: Plus Comma  : 0101111100 K28.5 (Normal)
-- Aurora 8/10: Minus Comma : 1010000011 K28.5 (Inverted)

signal rxctrl0 : slv(15 downto 0) := (others => '0');
signal rxctrl1 : slv(15 downto 0) := (others => '0');
signal rxctrl2 : slv(7 downto 0)  := (others => '0');
signal rxctrl3 : slv(7 downto 0)  := (others => '0');

-----------------------------------------------------
------- Custom Signals for TX Data generation -------
signal TRX_Cnt_Data_Max   : slv(31 downto 0) := slv(to_unsigned(256 - 1,32));
signal TRX_Cnt_Corr_Max   : slv(31 downto 0) := slv(to_unsigned(4 - 1  ,32));
signal TRX_Cnt_Comma_Max  : slv(31 downto 0) := slv(to_unsigned(16 - 1  ,32));
signal TRX_Data           : slv(31 downto 0) := (others => '0');
signal TRX_Cnt            : slv(31 downto 0) := (others => '0');
type TRX_State_t is (Transmit_Comma,Transmit_ClockCorrection,Transmit_Data);
signal TRX_State : TRX_State_t;

-------------------------------------------------
------- Auora Signals for Synchronization -------
type Aurora_t is (Aurora_Wait_Serdes,Aurora_Align_Positive,Aurora_Align_Wait,Aurora_Align_Negative,Aurora_Operational);
signal Aurora : Aurora_t;
signal Aurora_Byte_Aligned      : sl := '0';
signal Aurora_Comma_Detected    : sl := '0';
signal Aurora_CommaAlign_Minus  : sl := '0';
signal Aurora_CommaAlign_Plus   : sl := '0';
signal Aurora_BufReset          : sl := '0';
signal Aurora_Align_Changed     : sl := '0';
signal Aurora_Wait_Cnt          : slv(15 downto 0) := (others => '0');
signal Aurora_Wait_Cnt_Max      : slv(15 downto 0) := slv(to_unsigned(255,16));
signal Aurora_Buffer_Status     : slv(2 downto 0) := "000";
signal Aurora_Buff_Cnt          : slv(1 downto 0) := "00";

----------------------------------------------------------------------------
begin -------- -------- -------- Architecture begins ------- ------- -------
----------------------------------------------------------------------------

------------------------------------------------------------
-- This is 100 MHz for Our TB & Transceiver Configuration --
DRP_FreeClk_Gen:process
begin
  Clk_Free <= '0';
  wait for Period_Free / 2;
  Clk_Free <= '1';
  wait for Period_Free / 2;
end process;

------------------------------------------------------------
-- This is 200 MHz for Our TB & Transceiver Configuration --
MGT_RefClk_Gen:process
begin
  Clk_Ref <= '0';
  wait for Period_Ref / 2;
  Clk_Ref <= '1';
  wait for Period_Ref / 2;
end process;

TB_Stimuli:process
begin

  -------------------------------------------------------------
  -- Reset the Transceiver under normal operating conditions --
  -- RX Data should mirror the TX Data without any Disparity / NotInTable errors
  Loopback_Mode <= LOOP_NEAR_PMA;
  PRBS_Mode     <= PRBS_NORMAL;
  wait for 100 ns;
  Transceiver_Reset <= '1';
  RX_PMA_Reset      <= '1';
  TX_PMA_Reset      <= '1';
  wait for 10 ns;
  Transceiver_Reset <= '0';
  RX_PMA_Reset      <= '0';
  TX_PMA_Reset      <= '0';
  wait until (GT_TX_Reset_Done = '1');
  wait until (GT_RX_Reset_Done = '1');
  wait for 10 ns;
  RX_PMA_Reset      <= '1';
  wait for 10 ns;
  RX_PMA_Reset      <= '0';
  wait for 20 us;

  -----------------------------------------------------------
  -- Set UP PRBS on RX & TX and also Near-End PMA Loopback --
  -- Note: There will be Disparity/NotInTable Errors in this mode.
  -- Important is to look for PRBS_Locked & PRBS_Error signals.
  Loopback_Mode <= LOOP_NEAR_PMA;
  PRBS_Mode     <= PRBS_31;
  RX_PMA_Reset  <= '1';
  PRBS_Reset    <= '1';
  wait for 10 ns;
  RX_PMA_Reset <= '0';
  wait until (RX_PMA_Rst_Done = '1');
  PRBS_Reset <= '0';
  wait for 20 us;

  wait;
end process;

----------------------------------------------
-- Note: Data Transmissin is from LSBs to MSBs
-- IE: WORD = [ B3 B2 B1 B0] gets transmitted:
-- t0 t1 t2 t3
-- B0 B1 B2 B3
-- Note: Data Reception is from LSBs to MSBs
-- t0 t1 t2 t3
-- B0 B1 B2 B3
-- IE: WORD = [B3 B2 B1 B0]
----------------------------------------------
Simple_TX_Side:process(TX_Clk_Buffed)
  variable seed1        : positive := 22;
  variable seed2        : positive := 17;
  variable x            : real;
  variable random_num_0 : integer;
  variable random_num_1 : integer;
  variable random_num_2 : integer;
  variable V_DataWord_0 : slv(7 downto 0);
  variable V_DataWord_1 : slv(7 downto 0);
  variable V_DataWord_2 : slv(7 downto 0);
begin
   if Transceiver_Reset = '1' then
    TRX_State   <= Transmit_Comma;
    TX_Data     <= (others => '0');
    TX_CharIS_K <= (others => '0');
    TX_Disp_0   <= (others => '0');
    TX_Disp_1   <= (others => '0');
    TRX_Cnt     <= (others => '0');
    TRX_Data    <= (others => '0');
  elsif rising_edge(TX_Clk_Buffed) then
      case TRX_State is
        ---------------------------------------------------------------
        when Transmit_Comma => ------- Transmit Comma Alignment -------
          -------------------------------------------------------------
          -- RX_DATA_WIDTH      = 32/40
          -- RX_INT_DATA_WIDTH  = 1
          -- ALIGN_COMMA_WORD   = 4 (Alignment Boundary == 4 Byte Boundary)
          -- See UG579 "Comma Alignment Boundaries" Figure 4-27,P.245
          -- The Only Supported position for 4-Byte Boundarios is that K28.5 is at Byte 0.
          --------------------------------------------------------------

          -----------------------------------------------------------------------------
          -- Note: This should help with CDR (Clock Data Recovery Circuit) at the RX --
          -- It should be sufficient to Sent a free running counter insted or use
          -- pattern per the protocol specification such as D_10_2 & D_10_2 & D_10_2 & K_28_5
          uniform(seed1, seed2, x);
          random_num_0 := integer(floor(x * 256.0));
          uniform(seed1, seed2, x);
          random_num_1 := integer(floor(x * 256.0));
          uniform(seed1, seed2, x);
          random_num_2 := integer(floor(x * 256.0));
          V_DataWord_0 := slv(to_unsigned(random_num_0,8));
          V_DataWord_1 := slv(to_unsigned(random_num_1,8));
          V_DataWord_2 := slv(to_unsigned(random_num_2,8));


          TX_Data     <= V_DataWord_0 & V_DataWord_1 & V_DataWord_2 & K_28_5;
          TX_CharIS_K <= "0001";          -- Mark K28.5 Byte as Control Symbol
          TX_Disp_0   <= (others => '0'); -- Let Encoder Decide for disparity
          TX_Disp_1   <= (others => '0'); -- Let Encoder Decide for disparity
          
          if TRX_Cnt = TRX_Cnt_Comma_Max then
            TRX_Cnt     <= (others => '0');
            TRX_State   <= Transmit_ClockCorrection;
          else
            TRX_Cnt <= slv(unsigned(TRX_Cnt) + 1);
          end if;
        ------------------------------------------------------------------------------
        when Transmit_ClockCorrection => ------- Clock Correction Transmission -------
          --  --------------------------------------------------------------------------
          --  -- RX_DATA_WIDTH      = 32/40
          --  -- RX_INT_DATA_WIDTH  = 1
          --  -- ALIGN_COMMA_WORD   = 4 (Alignment Boundary == 4 Byte Boundary)
          --  -- See UG579 "Comma Alignment Boundaries" Figure 4-27,P.245
          --  --------------------------------------------------------------
          -- Note: clock Correction requires 4 Words as per the GTWIZ's configuration
          -- Note: clock correction sequence amount is bound to UG578 "Valid ALIGN_COMMA_WORD/CLK_COR_SEQ_LEN Combinations"
          TX_Data     <= K_23_7 & K_23_7 & K_23_7 & K_23_7;
          TX_CharIS_K <= "1111";          -- Mark Bytes as Control Symbols
          TX_Disp_0   <= (others => '0'); -- Let Encoder Decide for disparity
          TX_Disp_1   <= (others => '0'); -- Let Encoder Decide for disparity

          if TRX_Cnt = TRX_Cnt_Corr_Max then
            TRX_Cnt <= (others => '0');
            if Aurora = Aurora_Operational then
              TRX_State   <= Transmit_Data;
            else
              TRX_State   <= Transmit_Comma;
            end if;
          else
            TRX_Cnt <= slv(unsigned(TRX_Cnt) + 1);
          end if;

        ------------------------------------------------------------
        when Transmit_Data  =>  ------- Transmit Standard Data ------
          TX_Data     <= TRX_Data;
          TX_CharIS_K <= "0000";          -- Mark Bytes as standard Data
          TX_Disp_0   <= (others => '0'); -- Let Encoder Manage Disparity
          TX_Disp_1   <= (others => '0'); -- let Encoder Manage Disparity
          TRX_Data    <= slv(unsigned(TRX_Data) + 1);
          if TRX_Cnt = TRX_Cnt_Data_Max then
            TRX_State <= Transmit_Comma;
            TRX_Cnt   <= (others => '0');
          else
            TRX_Cnt <= slv(unsigned(TRX_Cnt) + 1);
          end if;
      end case;
  end if;
end process;



PolarLight:process(RX_Clk_Buffed)
begin
  if rising_edge(RX_Clk_Buffed) then
    if Transceiver_Reset = '1' then
      Aurora                  <= Aurora_Wait_Serdes;
      Aurora_CommaAlign_Minus <= '0';
      Aurora_CommaAlign_Plus  <= '0';
      Aurora_BufReset         <= '1';
      Aurora_Wait_Cnt         <= (others => '0');
    else
      ---------------------------------------------------------------
      ------ Disable Elastic Buffer Reset for Clock Correction ------
      -- Note: Under normal Circumstances,the FSM should control the buffer status
      -- Note: Based on Protoc,Buffer might be selected to by automatically reset upon comma reception
      Aurora_BufReset     <= '0';

      case Aurora is
        --------------------------------------------------------------------------------
        when Aurora_Wait_Serdes => ------- Wait Until Serdes Comes Out of Reset -------
          if GT_TX_Reset_Done = '1' and GT_RX_Reset_Done = '1' and CDR_Stable = '1' then
            Aurora <= Aurora_Align_Positive;
          end if;
        -----------------------------------------------------------------------------
        when Aurora_Align_Positive    => ------- Wait for Positive Comma Sync -------
          Aurora_CommaAlign_Plus  <= '1';
          if Aurora_Comma_Detected = '1' and Aurora_Byte_Aligned = '1' and Aurora_CommaAlign_Plus = '1' then
            Aurora_CommaAlign_Plus  <= '0';
            Aurora                  <= Aurora_Align_Wait ;
          end if;
        ----------------------------------------------------------------------------------------
        when Aurora_Align_Wait => ------- Wait a while before issuing negative Alignment -------
          if Aurora_Wait_Cnt = Aurora_Wait_Cnt_Max then
            Aurora_Wait_Cnt <= (others => '0');
            Aurora          <= Aurora_Align_Negative;
          else
            Aurora_Wait_Cnt <= slv(unsigned(Aurora_Wait_Cnt) + 1);
          end if;
        -----------------------------------------------------------------------------
        when Aurora_Align_Negative    => ------- Wait for Negative Comma Sync -------
          Aurora_CommaAlign_Minus   <= '1';
          if Aurora_Comma_Detected = '1' and Aurora_Byte_Aligned = '1' and Aurora_CommaAlign_Minus = '1' then
            Aurora_CommaAlign_Minus <= '0';
            Aurora_BufReset         <= '1';
            Aurora                  <= Aurora_Operational;
          end if;

        -------------------------------------------------------
        when Aurora_Operational  => ------- Operational -------
      end case;
    end if;
  end if;
end process;


------------------------------------------------------------------------
------ Make Sure that TX gets Reset first and then RX gets Reset -------
-- #0 Wait for PowerGood
-- #1 Deassert TX Reset and Wait for TX Reset to Complete
-- #2 Deassert RX Reset and Wait for RX Reset to complete
SimpleResetFSM:process(Clk_Free)
begin
  if rising_edge(Clk_Free) then
    if Transceiver_Reset = '1' then
      Transceiver_Reset_TX        <= '1';
      Transceiver_Reset_RX        <= '1';
      AuroraReset_FSM             <= WAIT_POWER_GOOD;
    else
      case AuroraReset_FSM is
        ------------------------------------------------------
        when WAIT_POWER_GOOD        =>  ----------------------
          if GT_PowerGood = '1' then
            Transceiver_Reset_TX  <= '0';
            AuroraReset_FSM       <= WAIT_TX_RESET_COMPLETE;
          end if;
        ------------------------------------------------------
        when WAIT_TX_RESET_COMPLETE =>  ----------------------
          if GT_TX_Reset_Done = '1' then
            Transceiver_Reset_RX  <= '0';
            AuroraReset_FSM       <= WAIT_RX_RESET_COMPLETE;
          end if;
        ------------------------------------------------------
        when WAIT_RX_RESET_COMPLETE => ----------------------
          if GT_RX_Reset_Done = '1' then
            AuroraReset_FSM <= RESET_DONE;
          end if;
        ------------------------------------------------------
        when RESET_DONE             => ----------------------
      end case;
    end if;
  end if;
end process;

---------------------------------------------------------------
-- GTY Transceiver at 4Gbit / Sec / Lane Targeted for VCU118 --
GT_Transceiver:Aurora_GTY
port map(
  gtwiz_userclk_tx_active_in(0)           => TX_Active,
  gtwiz_userclk_rx_active_in(0)           => RX_Active,
  gtwiz_reset_clk_freerun_in(0)           => Clk_Free,
  gtwiz_reset_all_in(0)                   => '0',
  gtwiz_reset_tx_pll_and_datapath_in(0)   => Transceiver_Reset_TX,
  gtwiz_reset_tx_datapath_in(0)           => '0',
  gtwiz_reset_rx_pll_and_datapath_in(0)   => Transceiver_Reset_RX,
  gtwiz_reset_rx_datapath_in(0)           => '0',
  gtwiz_reset_rx_cdr_stable_out(0)        => CDR_Stable,
  gtwiz_reset_tx_done_out(0)              => GT_TX_Reset_Done,
  gtwiz_reset_rx_done_out(0)              => GT_RX_Reset_Done,

  -------------------------------
  ------ SERDES DATA PORTS ------
  gtwiz_userdata_tx_in      => TX_Data,
  gtwiz_userdata_rx_out     => RX_Data,

  gtrefclk00_in(0)          => Clk_Ref,
  gtrefclk01_in(0)          => Clk_Ref,
  qpll0outclk_out           => open,
  qpll0outrefclk_out        => open,
  qpll1outclk_out           => open,
  qpll1outrefclk_out        => open,
  
  -----------------------------------------
  ------- Custom Functionality Setup ------
  -- Note: Check PRBS with LOOP_NEAR_PMA
  -- For Correct Decoding and Encoding, use LOOP_NEAR_PCS
  loopback_in           => Loopback_Mode,
  rxprbssel_in          => PRBS_Mode,
  txprbssel_in          => PRBS_Mode,

  rxpd_in               => PD_NORMAL,
  txpd_in               => PD_NORMAL,
  rx8b10ben_in(0)       => B8B10_ENABLED,
  tx8b10ben_in(0)       => B8B10_ENABLED,
  rxpolarity_in(0)      => POLARITY_NORMAL,
  txpolarity_in(0)      => POLARITY_NORMAL,
  rxpmareset_in(0)      => RX_PMA_Reset,
  txpmareset_in(0)      => TX_PMA_Reset,

  ------------------------------------
  -------- Driver Configuration ------
  txprecursor_in        => "00000", -- TX Pre  - Emphasis
  txpostcursor_in       => "00000", -- TX Post - Emphasis
  txdiffctrl_in         => "11111", -- Maximum TX Swing (Optionally)

  --------------------------
  ------- PRBS Status ------
  rxprbserr_out(0)      => PRBS_Error,
  rxprbslocked_out(0)   => PRBS_Locked,
  rxprbscntreset_in(0)  => PRBS_Reset,
  
  ----------------------------
  ------- User Clocking ------
  rxusrclk_in(0)           => RX_Clk_Buffed,
  rxusrclk2_in(0)          => RX_Clk_Buffed,
  txusrclk_in(0)           => TX_Clk_Buffed,
  txusrclk2_in(0)          => TX_Clk_Buffed,

  -----------------------------------
  ------- Serdes Differential ------- 
  gtyrxn_in(0)          => Serdes_Wire_N, -- N2N
  gtyrxp_in(0)          => Serdes_Wire_P, -- P2P
  gtytxn_out(0)         => Serdes_Wire_N, -- N2N
  gtytxp_out(0)         => Serdes_Wire_P, -- P2P

  --------------------------------------
  ------- Aurora Borealis Ports -------
  rxcommadeten_in(0)      => COMMA_ALIGN_ENABLED,       -- Enable Comma Alignment Block. Do Not Bypass
  rxmcommaalignen_in(0)   => Aurora_CommaAlign_Minus,   -- Enable Negative Comma Alignment
  rxpcommaalignen_in(0)   => Aurora_CommaAlign_Plus,    -- Enable Positive Comma Alignment

  rxbyteisaligned_out(0)  => Aurora_Byte_Aligned,   -- Byte Has been alligned
  rxbyterealign_out(0)    => Aurora_Align_Changed,  -- Active when Alignment Changes
  rxcommadet_out(0)       => Aurora_Comma_Detected, -- Comma Has Been Detected

  rxbufreset_in(0)        => Aurora_BufReset,       -- Resets Elastic Buffer
  rxbufstatus_out         => Aurora_Buffer_Status,  -- Elastic Buffer Status
  rxclkcorcnt_out         => Aurora_Buff_Cnt,       -- Elastic Buffer clock Correction


  ---------------------------------
  ------- RX / TX Registers -------
  rxctrl0_out           => rxctrl0,
  rxctrl1_out           => rxctrl1,
  rxctrl2_out           => rxctrl2,
  rxctrl3_out           => rxctrl3,

  -- Note: Running Disparity is calculated by the TX 8/10 encoder when 
  -- txctrl0 & txctrl1 are zeros. See UG578 "TXCTRL1 and TXCTRL0 versus Outgoing Disparity"
  -- TXCTRL0/1 [15:8] are unused when encoder is enabled.
  txctrl0_in            => slv(to_unsigned(0,8)) & "0000" & TX_Disp_0,    -- Optionally Force Running Disparity
  txctrl1_in            => slv(to_unsigned(0,8)) & "0000" & TX_Disp_1,    -- Optionally force Running Disparity
  txctrl2_in            => "0000" & TX_CharIS_K,                          -- Mark Bytes as control characters

  ----------------------------------
  ------- RX and TX Clocking -------
  rxoutclk_out(0)          => RX_Clk_Raw,
  txoutclk_out(0)          => TX_Clk_Raw,

  gtpowergood_out(0)       => GT_PowerGood,
  rxpmaresetdone_out(0)    => RX_PMA_Rst_Done,
  txpmaresetdone_out(0)    => TX_PMA_Rst_Done
);

-------------------------------------------------------------------------------
-- I guess this is rather not a part of the GTWIz's sizing helper module ... --
RX_CharIS_K       <= rxctrl0(3 downto 0);
RX_Disp_Err       <= rxctrl1(3 downto 0);
RX_CharIS_Comma   <= rxctrl2(3 downto 0);
RX_NotInTable     <= rxctrl3(3 downto 0);

--------------------------------------------------------------------
------ ------- User Clocking Helper Block for RX & TX ------ -------
-- Note: For 32/40-Bit ,the RXUSRCLK == RXUSRCLK2, This is based on 
-- UG578 "RXUSRCLK2 Frequency Relationship to RXUSRCLK" and Datawidth.
-- If USRCLK2 has to be half of the USRCLK,then different BUFG_GT "DIV"
-- settings shall be used.
-- Note: Each BUFG_GT should have BUFG_GT_SYNC unisim component.
-- This is automatically inserted by the vivado tool however.
RX_BUFG:BUFG_GT
 port map(
      CE      => '1',
      CEMASK  => '0',
      CLR     => Transceiver_Reset,
      CLRMASK => '0',
      DIV     => "000",
      I       => RX_Clk_Raw,
      O       => RX_Clk_Buffed
    );
TX_BUFG:BUFG_GT
 port map(
      CE      => '1',
      CEMASK  => '0',
      CLR     => Transceiver_Reset,
      CLRMASK => '0',
      DIV     => "000",
      I       => TX_Clk_Raw,
      O       => TX_Clk_Buffed
    );

------------------------------------------------------------------
-- Note: There should be Async_Reg Properties for Synth. Design --
TX_Active_Resync:process(TX_Clk_Buffed)
begin
  if Transceiver_Reset = '1' then
      TX_Clk_Active_Meta <= '0';
      TX_Clk_Active_Sync <= '0';
  elsif rising_edge(TX_Clk_Buffed) then
    TX_Clk_Active_Meta <= '1';
    TX_Clk_Active_Sync <= TX_Clk_Active_Meta;
  end if;
end process;

RX_Active_Resync:process(RX_Clk_Buffed)
begin
  if Transceiver_Reset = '1' then
      RX_Clk_Active_Meta <= '0';
      RX_Clk_Active_Sync <= '0';
  elsif rising_edge(RX_Clk_Buffed) then
    RX_Clk_Active_Meta <= '1';
    RX_Clk_Active_Sync <= RX_Clk_Active_Meta;
  end if;
end process;

TX_Active <= TX_Clk_Active_Sync;
RX_Active <= RX_Clk_Active_Sync;



end Behavioral;
